#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64
#include "stdio.h"
#include "stdlib.h"
#include "errno.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>           
#include <ext2fs/ext2_fs.h>
#include "unistd.h"
#include "journal.h"
#include "birra.h"

int hex_dump(char *buf, int size);
int super_block_dump(struct ext2_super_block sb);
off_t from_block_to_offset(int block);
char *read_blocks(int block, int size);

#define JFS_MAGIC_NUMBER    0xc03b3998U 
#define SUPER_BLOCK_OFFSET  1024
#define SUPER_BLOCK_SIZE    sizeof(struct ext2_super_block)
#define GROUP_DESC_SIZE     sizeof(struct ext2_group_desc)

typedef unsigned long bitmap_t;

int wrote_bytes;
int bitmap_size;
int device_fd;
int bytes_r;
int groups_n;
int dir_blocks_n = 0;
int allocated_groups;

FILE *file_del_in;
char *journal;

struct ext2_super_block super_block_s;
struct ext2_group_desc *group_desc_p;

typedef struct {
struct ext2_inode   *inodes_p;
bitmap_t            *inodes_bitmap_p;
bitmap_t            *block_bitmap_p;
} info_groups;

typedef struct {
int	inode;
int	parent;
int	current;
char    name[256];
} info_inodes;

info_inodes   info_inodes_s;
info_groups *info_groups_p;
journal_superblock_t *journal_superblock;

char *dir_blocks;

char output_path[1024];
char temp_path[2048];

#define INDIRECT_BLOCKS     EXT2_BLOCK_SIZE(&super_block_s) / 4

int main (int argc, char *argv[])
{
	int i;
	off_t offset;

	if (argc < 3)
	{
		printf("pass 3 parameters...PLEASE!!! ext3birra [device] [outputdir] \n ");
		exit(1);
	}
		
	device_fd = open(argv[1],O_RDONLY);

	if (device_fd == -1)
	{
		printf ("cannot open device... \n ");
		exit(1);
	}

	strcpy(output_path, argv[2]);

    i = strlen(output_path);

    if (output_path[i - 1] != '/')
    {
        output_path[i - 1] = '/';
        output_path[i] = 0;
    }

	if (mkdir_birra(output_path))
	{
		printf("outpur directory error \n");
		exit(1);
	}


	if (lseek(device_fd, 1024, SEEK_SET) == -1)	// 1024 = SB offset
	{
		printf("cannot lseek device \n");
		exit(1);
	}

	bytes_r = read(device_fd, &super_block_s, SUPER_BLOCK_SIZE); 	

	if (bytes_r == -1)
	{
		printf ("cannot read device \n ");
		exit(1);
	}

	super_block_dump(super_block_s);

	dir_blocks = malloc(super_block_s.s_blocks_count);

	if (dir_blocks == NULL)
	{
		printf("cannot allocate dir_blocks \n ");
		exit(1);
	}

//	hex_dump(&super_block_s, sizeof(struct ext2_super_block));

	/* calcolo il blocco del superblock */
	int super_block_block = SUPER_BLOCK_OFFSET / EXT2_BLOCK_SIZE(&super_block_s);

	if (EXT2_INODES_PER_GROUP(&super_block_s) == 0)
	{
		printf("inodes per group = 0 \n");
		exit(1);
	}

	/* numero totale gruppi presenti */
	groups_n = super_block_s.s_inodes_count / EXT2_INODES_PER_GROUP(&super_block_s);

	group_desc_p = calloc(groups_n,GROUP_DESC_SIZE);

	if (group_desc_p == NULL)
	{
		printf("error calloc group desc \n ");
		exit(1);
	}

	/* so che il descrittore di blocchi e il successivo al superblocck */
	int group_desc_block = super_block_block + 1;

	/* trovo l'offset del blocco */
	int bytes_to_read;

	bytes_to_read = groups_n * GROUP_DESC_SIZE;

	offset = from_block_to_offset(group_desc_block);

	if (lseek(device_fd, offset, SEEK_SET) == -1)
	{
		printf("error searching group desc block \n");
		exit(1);
	}

	bytes_r = read(device_fd, group_desc_p , bytes_to_read);
	
	if (bytes_r <= 0)
	{
		printf("error reading group desc block \n");
		exit(1); 
	}

	/* inizializzo i gruppi ma li fillo solo su richiesta */
	info_groups_p = calloc(groups_n,sizeof(info_groups));

	memset(info_groups_p, 0, (groups_n * sizeof(info_groups)));

	/* setto globalone per allocazioni */
	bitmap_size = EXT2_BLOCK_SIZE(&super_block_s) / sizeof(bitmap_t);


	/**************************************/
	/* Leggo il superblock journal        */
	/**************************************/

	char *journal_dev_p = NULL;

	struct ext2_inode *journal_inode_p;

	if (from_inode_number_to_inode((int) super_block_s.s_journal_inum, &journal_inode_p))
	{
		printf("error from_inode_number_to_inode \n");	
		exit (1);
	}

	char journal_buf[sizeof(journal_superblock_t)];
	char buf[EXT2_BLOCK_SIZE(&super_block_s)];
	FILE *fp;

	sprintf(temp_path, "%s/tmp/", output_path);
	mkdir_birra(temp_path);

	strcat(temp_path,"journal");

	
	if ((fp = fopen(temp_path,"w+")) == NULL)
	{
		printf("error dumping journal in %s \n ",temp_path);
		exit(1);
	}

	if (load_inode_file(journal_inode_p, fp, journal_inode_p->i_blocks))
	{
		printf("error load_inode data \n ");
		exit (1);
	}

	if (load_file(fp, &journal, (journal_inode_p->i_blocks * 512)))
	{
		printf("error loading journal in memory\n");
		exit(1);
	}

	journal_superblock = (journal_superblock_t *) journal;

	journal_superblock_dump(journal_superblock);

	if (be32_to_cpu(journal_superblock->s_header.h_magic) != JFS_MAGIC_NUMBER)
	{
		printf("magic number MAFFISSIMO %x not %x \n ",journal_superblock->s_header.h_magic, JFS_MAGIC_NUMBER);
		exit(1);
	}


	/**************************************/
	/* Esperimentone 		      */
	/**************************************/
	/* estraggo tutti i blocchi contenenti directories */

	printf("Disk analysis starting - Please wait some minute...\n ");

	scan_dir(EXT2_ROOT_INO, 0, EXT2_ROOT_INO);
	
	/**************************************/
	/* Alla ricerca dei fail perduti      */
	/**************************************/

	struct ext2_inode *inode_p;
	int search_blocknr, bit;			

/*	for(i=1;i<super_block_s.s_inodes_count;i++)
	{
		if (from_inode_number_to_inode(i, &inode_p))
		{
			printf("error from_inode_number_to_inode \n");	
			exit (1);
		}

		if (inode_p->i_dtime != 0)
		{

			if (from_inode_number_to_block(i, &search_blocknr, &bit))
			{
				printf("non trovo il blocco del inode %d =//// \n");
				continue;
			}	

			info_inodes_s.inode = i;
			info_inodes_s.parent = 0;
			info_inodes_s.current = 0;
			info_inodes_s.name[0] = 0;
			
			search_block_journal(search_blocknr, bit, journal, i);
		}

	}*/

	printf("Disk analysis OK - Starting recover \n ");

    if (file_del_in == NULL)
    {
        printf("no deleted data found =( \n ");
        exit(1);
    }

	fseek(file_del_in, 0L, SEEK_SET);

	while (!(feof(file_del_in)))
	{
		fread(&info_inodes_s, sizeof(info_inodes), 1, file_del_in);
				
		if (from_inode_number_to_inode(info_inodes_s.inode, &inode_p))
		{
			printf("error from_inode_number_to_inode  %d \n",info_inodes_s.inode);	
			continue;
		}

		if (inode_p->i_dtime != 0)
		{
			if (from_inode_number_to_block(info_inodes_s.inode, &search_blocknr, &bit))
			{
				printf("cannot fine inode block =//// \n");
				continue;
			}	

			search_block_journal(search_blocknr, bit, journal, info_inodes_s);
		} 
	}

		


	exit(0);
}



int hex_dump(char *buf, int size)
{
	int i;

	for (i=0; i<size; i+=2)
	{
		if (i % 16 == 0)
			printf("\n %04x    ",i);
		printf (" %04x",(unsigned short) buf[i]);
	}

	printf ("\n");
	
	return 0;
}

int super_block_dump(struct ext2_super_block sb)
{
	printf("*************** superblock *********************\n");
	printf(" s_inodes_count    : %d \n",sb.s_inodes_count);
	printf(" s_blocks_count    : %d \n",sb.s_blocks_count);
	printf(" s_log_block_size  : %d \n",sb.s_log_block_size);
	printf(" s_mtime           : %d \n",sb.s_mtime);
	printf(" s_state           : %d \n",sb.s_state);
	printf(" s_errors          : %d \n",sb.s_errors);
	printf(" s_first_inode     : %d \n",sb.s_first_ino);
	printf(" s_journal_inum    : %d \n",sb.s_journal_inum);
	printf(" s_journal_dev     : %d \n",sb.s_journal_dev);
	printf(" s_volume_name     : %s \n",sb.s_volume_name);

	return 0;
}

int group_desc_dump(struct ext2_group_desc *gd)
{
	printf(" bg_used_dirs_count: %d \n",gd->bg_used_dirs_count);

	return 0;
}

int journal_header_dump(journal_header_t *jh)
{	

	printf("****** journal header dump *********** \n ");
	printf(" magic 	     : %x \n",be32_to_cpu(jh->h_magic));
	printf(" sequence    : %d \n",be32_to_cpu(jh->h_sequence));

        switch (be32_to_cpu(jh->h_blocktype))
	{
                case JFS_DESCRIPTOR_BLOCK:
			printf(" blocktype    : DESCRIPTOR \n");
                        break;

                case JFS_COMMIT_BLOCK:
			printf(" blocktype    : COMMIT \n");
                        break;

                case JFS_REVOKE_BLOCK:
			printf(" blocktype    : REVOKE \n");
                        break;

		default:
			printf(" blocktype    : UNK \n");
			
	}

	return 0;
}

int journal_block_tag_dump(journal_block_tag_t *jbt)
{	

	printf("****** journal block_tag_dump *********** \n ");
	printf(" blocknr     : %d \n",be32_to_cpu(jbt->t_blocknr));
	printf(" flags       : %x \n",be32_to_cpu(jbt->t_flags));

	return 0;
}

int journal_superblock_dump(journal_superblock_t *js)
{
	printf("****** journal superblock dump *********** \n ");
	printf(" magic  	     : %x \n",be32_to_cpu(js->s_header.h_magic));
	printf(" transaction 	     : %d \n",be32_to_cpu(js->s_sequence));
	printf(" start	 	     : %d \n",be32_to_cpu(js->s_start));
	printf(" first	 	     : %d \n",be32_to_cpu(js->s_first));
	printf(" maxlen	 	     : %d \n",be32_to_cpu(js->s_maxlen));
	printf(" blocksize 	     : %d \n",be32_to_cpu(js->s_blocksize));

	return 0;
}

int dir_entry_dump(char *de)
{
	char name[300];
	int name_len;

	char *fake_p;

	struct ext2_dir_entry_2 *de_t;
	int shift;
	
	de_t = (struct ext2_dir_entry_2 *)de;

	while (((int)de_t < ((int)de + 4096)) && (de_t->rec_len != 0))
	{
		name_len = de_t->name_len & 0x00ff;
		
		if (name_len > 300)
			name_len = 300;

		memcpy(name, de_t->name, name_len);
		name[name_len] = 0;

		printf(" inode    : %d \n",de_t->inode);
		printf(" name     : %s \n",name);
		printf(" name_len : %d \n",de_t->name_len & 0x00ff);
		printf(" rec_len  : %d \n",de_t->rec_len);
		printf(" inode    : %d \n",de_t->inode);

		shift = (unsigned long) de_t->rec_len;
	
		printf("prim %d + %d \n",de_t,shift);

		fake_p = (char *) de_t;
		de_t = &fake_p[shift];

		printf("ora punto a %d \n",de_t);

	}

	return 0;
}


int inode_dump(struct ext2_inode *in)
{
	int i;

	if ((in->i_mode & 0xf000) == 0x4000)
		printf(" DIRECTORY \n");

	printf(" i_mode : %d \n",in->i_mode);
	printf(" i_mode sab : %d \n",(in->i_mode & 0xf000) >> 24);
	printf(" i_atime: %d \n",in->i_atime);
	printf(" i_dtime: %d \n",in->i_dtime);
	printf(" i_uid  : %d \n",in->i_uid);
	printf(" i_size  : %d \n",in->i_size);
	printf(" i_blocks  : %d \n",in->i_blocks);

	if (in->i_blocks == 0 && in->i_size != 0)
	{
		printf("RIFLETTI DVDN..... \n");
//		sleep (5);
	}

	return 0;
}

off_t from_block_to_offset(int block)
{
	off_t offset;
	off_t temp;

	temp = block;
	offset = temp * EXT2_BLOCK_SIZE(&super_block_s);

	return offset;

}

int from_inode_to_dir(struct ext2_inode *in, struct ext2_dir_entry_2 **dir)
{
	unsigned int i;
	off_t offset;
	char *temp;
	int blocks;

	blocks = in->i_blocks;

	*dir = malloc(blocks * 512);

	if (*dir == NULL)
		return(1);

	temp = (char *) *dir;

	blocks /= (EXT2_BLOCK_SIZE(&super_block_s) / 512);

	for (i=0; i<blocks; i++)
	{
		/* TODO */
		if (i == 12)
			break;

		/* questo blocco e una dir */
		dir_blocks[in->i_block[i]] = 1;

		dir_blocks_n++;

		offset = from_block_to_offset((int) in->i_block[i]);
	
		if (lseek(device_fd, offset, SEEK_SET) == -1)
		{
			printf("cannot seek %lld \n ", offset);
			free(*dir);
			return (1);
		}

		bytes_r = read(device_fd, &temp[i * EXT2_BLOCK_SIZE(&super_block_s)], EXT2_BLOCK_SIZE(&super_block_s)); 

	}

	
	return 0;
}


int load_block_bitmap (bitmap_t **bb, int group)
{
/*	int *fake;

	fake = malloc(bitmap_size);

	return (0); */

	*bb = malloc(bitmap_size);

	if (*bb == NULL)
		return(1);

	off_t offset = from_block_to_offset(group_desc_p[group].bg_block_bitmap);

	if (lseek(device_fd, offset, SEEK_SET) == -1)
	{
		printf(" BAD offset %lld - errno %d \n",offset,errno);
		return 1;
	}

		
	bytes_r = read(device_fd, *bb, bitmap_size); 	

	if (bytes_r <= 0)
		return 1;
	
	return 0;	
}	


int load_inodes_bitmap (bitmap_t **bb, int group)
{

	*bb = malloc(bitmap_size);

	if (*bb == NULL)
	{
		printf("error malloc LOAD INODE BITMAP \n ");
		return 1;
	}

	off_t offset = from_block_to_offset(group_desc_p[group].bg_inode_bitmap);

	if (lseek(device_fd, offset, SEEK_SET) == -1)
		return 1;
		
	bytes_r = read(device_fd, *bb, bitmap_size); 	

	if (bytes_r <= 0)
		return 1;

	return 0;	
}	


int load_inodes_group (struct ext2_inode **in, int group)
{
	*in = malloc(EXT2_INODES_PER_GROUP(&super_block_s) * sizeof(struct ext2_inode));

	if (*in == NULL)
		return(1);

	if (group > groups_n)
	{
		free(*in);
		return(1);
	} 

	off_t offset = from_block_to_offset(group_desc_p[group].bg_inode_table);

	if (lseek(device_fd, offset, SEEK_SET) == -1)
		return (1);
		
	bytes_r = read(device_fd, *in, EXT2_INODES_PER_GROUP(&super_block_s) * sizeof(struct ext2_inode)); 	

	if (bytes_r <= 0)
		return 1;

	return 0;	
}	


int from_inode_number_to_inode(int in_n, struct ext2_inode **in)
{
	int i;
	unsigned int bit;
	int blocknr;

/*	if (!(inode_is_allocated(in_n)))
	{
		if (from_inode_number_to_block(in_n, &blocknr, &bit))
			return(1);

		if (get_inode_from_journal(blocknr, bit, journal, in))
			return(1);
		else
			return(0);
	}	*/

	int group = (in_n - 1) / EXT2_INODES_PER_GROUP(&super_block_s);

	bit = (in_n - 1) - (group * EXT2_INODES_PER_GROUP(&super_block_s));

	if (group > groups_n)
		return(1);

	if (info_groups_p[group].inodes_p == 0)
	{
		if (allocated_groups > 30) 
		{
			for (i=0; (i<group && i<groups_n); i++)
			{
				if (info_groups_p[i].inodes_p != NULL)
				{
					free(info_groups_p[i].inodes_p);
					info_groups_p[i].inodes_p = 0;
					allocated_groups--;
				}
			}
		}

		if (load_group_inodes(group))
			return 1;
	}

	*in = &info_groups_p[group].inodes_p[bit];

	return 0;
}	


int from_inode_number_to_block(int in_n, int *search_blocknr, int *bit)
{

	int group = (in_n - 1) / EXT2_INODES_PER_GROUP(&super_block_s);

	unsigned int temp_bit = (in_n - 1) - (group * EXT2_INODES_PER_GROUP(&super_block_s));

	if (group > groups_n)
		return (1);

	int group_block = group_desc_p[group].bg_inode_table;

	int temp = temp_bit * sizeof(struct ext2_inode);

	int temp_blocknr = temp / EXT2_BLOCK_SIZE(&super_block_s);

	*bit = temp % EXT2_BLOCK_SIZE(&super_block_s);

	*search_blocknr = temp_blocknr + group_block;

	return 0;
}	


int load_group_inodes(int g)
{
	if (load_inodes_group(&info_groups_p[g].inodes_p, g))
		return 1;
	if (load_inodes_bitmap(&info_groups_p[g].inodes_bitmap_p, g))
		return 1;


	allocated_groups++;

	return 0;
}

int load_inode_file_indirect(int indirect_block, FILE *fp, int max, int *ind, int level, int max_bytes)
{
	char buf[EXT2_BLOCK_SIZE(&super_block_s)];
	int indirect_buf[INDIRECT_BLOCKS];
	int i;
	
	off_t offset = from_block_to_offset(indirect_block);

	if (lseek(device_fd, offset, SEEK_SET) == -1)
	{
		printf("error trying seeking inode data - off %lld blk %d \n",offset,indirect_block);
		return (1);
	}
	
	if ((bytes_r = read(device_fd, indirect_buf, EXT2_BLOCK_SIZE(&super_block_s))) <= 0)
	{
		printf("error trying reading indirect block - %s \n ",strerror(errno) );
		return (1);
	}
	
	int max_temp = *ind;

	for (i=0; i<INDIRECT_BLOCKS; i++, max_temp++)
	{

		if (indirect_buf[i] == 0)
			break;

		if (max_temp >= max)
		{
			printf("ind > max  - %d %d\n ",max_temp,max);
			break;
		}

		if (level > 1)
		{
			if (load_inode_file_indirect(indirect_buf[i], fp, max, ind, level - 1, max_bytes))
				return (1);

			max_temp = *ind;

			if (max_temp >= max)
				return (0);

			continue;
		}


		offset = from_block_to_offset(indirect_buf[i]);

		if (lseek(device_fd, offset, SEEK_SET) == -1)
		{
			printf("error trying seeking inode data - offset %lld blocke %d \n",offset,indirect_buf[i]);
			return (1);
		}

		int temp = max_temp * EXT2_BLOCK_SIZE(&super_block_s);

		if ((bytes_r = read(device_fd, buf , EXT2_BLOCK_SIZE(&super_block_s))) <= 0)
		{
			printf("error trying reading indirect block data - %s \n ",strerror(errno) );
			return (1);
		}

		wrote_bytes += EXT2_BLOCK_SIZE(&super_block_s);

		int bytes_to_write;

		if (wrote_bytes > max_bytes)
		{
			bytes_to_write = EXT2_BLOCK_SIZE(&super_block_s) - (wrote_bytes - max_bytes);

			wrote_bytes = max_bytes;
		}
		else
			bytes_to_write = EXT2_BLOCK_SIZE(&super_block_s);

		if (bytes_to_write == 0)
			continue;

		if (fwrite(buf, bytes_to_write, 1, fp) == 0)
		{
			printf("error trying writing block data\n");
			return(1);
		}
	
		
	}

	*ind = max_temp + 1;
	
	return(0);
}	


int load_inode_file(struct ext2_inode *in, FILE *fp, int max_blocks)
{
	unsigned int i;
	int max_bytes;
	off_t offset;
	char buf[EXT2_BLOCK_SIZE(&super_block_s)];

	wrote_bytes = 0;
	
	if (max_blocks == 0)
		max_bytes = in->i_size;
	else
		max_bytes = max_blocks * EXT2_BLOCK_SIZE(&super_block_s);

	max_blocks = in->i_blocks / (EXT2_BLOCK_SIZE(&super_block_s) / 512) ;

	for (i=0; i< max_blocks; i++)
	{
		if (i == 12)
		{
			if (load_inode_file_indirect(in->i_block[i], fp, max_blocks, &i, 1, max_bytes))
				return(1);
			
			if (i >= max_blocks)
				break;

			continue;
		}

		if (i == (INDIRECT_BLOCKS + 12 + 2))
		{
			if (load_inode_file_indirect(in->i_block[13], fp, max_blocks, &i, 2, max_bytes))
				return(1);

			if (i >= max_blocks)
				break;


			continue;
		}

 		offset = from_block_to_offset((int) in->i_block[i]);

		if (lseek(device_fd, offset, SEEK_SET) == -1)
		{
			printf("error trying seeking inode data - off %lld blk %d \n",offset,in->i_block[i]);
			return (1);
		}

		if ((bytes_r = read(device_fd, buf, EXT2_BLOCK_SIZE(&super_block_s))) <= 0)
		{
			printf("error trying reading inode data - %s \n ",strerror(errno) );
			return (1);
		}

		wrote_bytes += EXT2_BLOCK_SIZE(&super_block_s);

		int bytes_to_write;

		if (wrote_bytes > max_bytes)
		{
			bytes_to_write = EXT2_BLOCK_SIZE(&super_block_s) - (wrote_bytes - max_bytes);

			wrote_bytes = max_bytes;
		}
		else
			bytes_to_write = EXT2_BLOCK_SIZE(&super_block_s);


		if (bytes_to_write == 0)
			continue;
 
		if (fwrite(buf, bytes_to_write, 1, fp) == 0)
		{
			printf("error trying writing block data\n");
			return(1);
		}
		
	}

	return 0;
}


char *read_blocks(int block, int size)
{
	char *buf;
	off_t offset;

	offset = from_block_to_offset(block);

	if (lseek(device_fd, offset, SEEK_SET) == -1)
		return (0);

	if ((buf = malloc(size)) == NULL)
		return (0);

	if (read(device_fd, buf, size) <= 0)
		return (0);

	return buf;
}


int loop_descriptor_block(int *blocknr, journal_header_t *jh, char *journal, int blocksize, int search_blocknr, int inode_offset, info_inodes iis)
{
	journal_block_tag_t *jbt;
	int offset = sizeof(journal_header_t);
	char *temp = (char *) jh;
	int tag_block, tag_flags;
	int blocknr_temp;
	int temp_offset, i;
	int device_offset;
	int temp_inode;
	int name_len;
	char name[300];
	char *temp_p;
	int parent,current;
	char filename[700];
	char path[700];
	int shift;

	struct ext2_inode *in, *last_good_in = NULL;

	blocknr_temp = *blocknr;

	memset(name,0,300);
	
	

	++blocknr_temp;
	WRAP(journal_superblock, blocknr_temp);

	struct ext2_dir_entry_2 *dir = NULL;

	parent = 0;

	do  	
	{

		jbt = &temp[offset];

		offset += sizeof(journal_block_tag_t);

		if (offset > blocksize)
			break;
		
		tag_flags = be32_to_cpu(jbt->t_flags);		
		tag_block = be32_to_cpu(jbt->t_blocknr);		

/*		if (is_inode
		if (dir_blocks[tag_block])
		{
			struct ext2_dir_entry_2 *dir;

			temp_offset = blocknr_temp * 4096;;

			dir = &journal[temp_offset];

			int block_offset = 0;

			while (!(check_dir(dir)))
			{
				set_iis(dir, &iis);

				shift = 8 + dir->name_len;
				
				while (shift % 4)
					shift++;

				if (dir->rec_len != shift)
				{
					int final_offset = dir->rec_len;

					while (shift < final_offset)
					{
						dir = &journal[temp_offset + block_offset + shift];

						if (!(check_dir(dir)))
							set_iis(dir, &iis);

						shift += 4;
					}

					block_offset += final_offset;
				}
				else
					block_offset += shift;
							
				dir = &journal[temp_offset + block_offset];
			}

		}*/

		if (tag_block == search_blocknr)
		{
		//	printf("faundato qualcosa relativo al blocco %d in pos %d \n ",tag_block,blocknr_temp);
		//	printf("vado a ravanare dentro - %d  \n ",inode_offset);

		//	inode_dump(in);

			temp_offset = blocknr_temp * blocksize;

			temp_offset += inode_offset;

			in = &journal[temp_offset];

			if (((in->i_dtime == 0) && ((in->i_mode & 0xf000) == 0x8000)) && (in->i_size > 0))
			{
				last_good_in = in;
			}

		}
			

		if (tag_flags & JFS_FLAG_DELETED)
			printf("il blocco %d e stato cancellato \n");

//		read_blocks(tag_block, blocksize);
			
                if (!(tag_flags & JFS_FLAG_SAME_UUID))
                        offset += 16;

		++blocknr_temp;
		WRAP(journal_superblock, blocknr_temp);

	} while (!(tag_flags & JFS_FLAG_LAST_TAG));

	if (last_good_in)
	{

	memset(filename,0,700);
	memset(path,0,700);

/*	if (fseek(file_del_in, 0L, SEEK_SET) == -1)
		printf("error seeking file /tmp/file_del \n ");
	else
	{
		while (!feof(file_del_in))
		{
			fread(&info_inodes_s, sizeof(info_inodes), 1, file_del_in);

			if (info_inodes_s.inode == inode)
			{
				printf("TROVATOOOOOOOOOOOOOOOOOO \n");
				current = info_inodes_s.current;
				parent = info_inodes_s.parent;
				strcpy(name, info_inodes_s.name);
				break;
			}
		}
	}*/

	if (iis.parent != 0)
	{
		from_inode_number_to_path(iis.parent, iis.current, path);

	}

	if (strlen(iis.name) == 0)
		sprintf(name,"inode%d",iis.inode);
	else
		strcpy(name, iis.name);

	sprintf(filename, output_path);

	if (path[0] != 0)
	{
			
		strcat(filename, path);

		i = strlen(filename);

		while (filename[i] != '/')	
			i--;

		filename[i] = 0;

		printf("creating DIR %s \n ",filename);

		if (mkdir_birra(filename))
		{
			*blocknr = blocknr_temp;
			return (1);
		}

		filename[i] = '/';

	}

	strcat(filename,name);

	printf("writing file %s \n ",filename);
	
	FILE *fp = fopen(filename,"w+");

	if (fp != NULL)
	{
		if (load_inode_file(last_good_in, fp, 0))
			printf("ERRORE NEL RESCHIU \n");
	
		fclose(fp);
	}
	
		

	}

	*blocknr = blocknr_temp;

	if (dir)
		free(dir);
}

int search_block_journal(int search_blocknr, int inode_offset, char *journal, info_inodes iis)
{
	journal_superblock_t *journal_superblock;
	journal_block_tag_t *journal_block_tag;
	journal_header_t *journal_header;

	journal_superblock = (journal_superblock_t *) journal;

	int blocksize = be32_to_cpu(journal_superblock->s_blocksize);
	int blocknr = be32_to_cpu(journal_superblock->s_start);
	int transaction = be32_to_cpu(journal_superblock->s_sequence);

	while (1)
	{
		journal_header = (journal_header_t *) &journal[blocknr * blocksize];

		if (journal_header == NULL)
		{
			printf("error reading journal \n");
			exit(1);
		}

		if ((be32_to_cpu(journal_header->h_sequence) != transaction) || (be32_to_cpu(journal_header->h_magic) != JFS_MAGIC_NUMBER))
			break;


//		journal_header_dump(journal_header);

/*		journal_block_tag = &journal_header[sizeof(journal_header)];

		journal_block_tag_dump(journal_block_tag); */

	        switch (be32_to_cpu(journal_header->h_blocktype))
        	{
                	case JFS_DESCRIPTOR_BLOCK:
					loop_descriptor_block(&blocknr, journal_header, journal, be32_to_cpu(journal_superblock->s_blocksize), search_blocknr, inode_offset, iis);	
                        	break;

                	case JFS_COMMIT_BLOCK:
				blocknr++;
				transaction++;
                        	break;

                	case JFS_REVOKE_BLOCK:
				blocknr++;
                        	break;

                	default:
                        	printf(" blocktype    : UNK \n");
        	}

	}	

	return 0;
}
		
			

int load_file(FILE *fp, char **buf, int size)
{

	*buf = malloc(size);

	if (*buf == NULL)
		return(1);

	if (fseek(fp, 0L, SEEK_SET) == -1)
	{
		free(*buf);
		return(1);
	}

	char *temp;
	
	temp = *buf;
	while (fread(temp, 512, 1, fp) > 0)
		temp += 512;

	return (0);
}
	



int scan_dir(int in_n, int level, int parent)
{
	struct ext2_inode *inode;
	char *dir;
	struct ext2_dir_entry_2 *single_dir, *single_dir_del;
	int i, len;
	int real_offset = 0;
	char name [1024];

	if (from_inode_number_to_inode(in_n, &inode))
		return(1);

	if (from_inode_to_dir(inode, (struct ext2_dir_entry_2 **) &dir))
		return(1);

	// free(inode);

	i = 0;

	int inode_size = inode->i_size;
	int z = 0;

	while(i < inode_size)
	{
		single_dir = (struct ext2_dir_entry_2 *) &dir[real_offset];

		if ((single_dir->name_len == 0) || (single_dir->inode == 0))
			break;

		if (from_inode_number_to_inode(single_dir->inode, &inode))
		{
			break;
		}
		
		len = single_dir->name_len;

		if ((inode->i_mode & 0xf000) == 0x8000)
		{
			memcpy(name, single_dir->name, len);
			name[len] = 0;
		}
		
		if ((inode->i_mode & 0xf000) == 0x4000)
		{
			if ((single_dir->name_len != 1) && (single_dir->name[0] != '.'))
			{
				if ((single_dir->name[0] == '.') && (single_dir->name[1] == '.') && (single_dir->name_len == 2))
					printf("=");
				else
				{
				memcpy(name, single_dir->name, len);
				name[len] = 0;

				scan_dir(single_dir->inode, level + 1, in_n);
				}
			}
		}

		int shift = 8 + single_dir->name_len;

		while (shift % 4)
			shift++;
		
		if (single_dir->rec_len != shift)
		{
			int final_offset = real_offset + single_dir->rec_len;
			i = real_offset + shift;

			while (i < final_offset)
			{
				single_dir_del = (struct ext2_dir_entry_2 *) &dir[i];
				
				if (!(check_dir(single_dir_del)))
				{
					if (file_del_in == NULL)
					{
						sprintf(temp_path, "%s/tmp/", output_path);
						mkdir_birra(temp_path);
						strcat(temp_path,"file_del");

						file_del_in = fopen(temp_path, "w+");
		
						if (file_del_in == NULL)
						{
							printf("error opening file file_del\n");
							exit(1);
						}
					}

					info_inodes_s.parent = parent;
					info_inodes_s.current = in_n;
					info_inodes_s.inode = single_dir_del->inode;
					len = single_dir_del->name_len;
					memcpy(info_inodes_s.name, single_dir_del->name, len);
					info_inodes_s.name[len] = 0;

					if (fwrite(&info_inodes_s, sizeof(info_inodes), 1, file_del_in) <= 0)
						printf("error writing file file_del \n ");
				
/*					printf(" verifico se è allocheto \n ");
	
					if (!(inode_is_allocated(single_dir_del->inode)))
					{
						if (from_inode_number_to_inode(single_dir_del->inode, &inode) == 0)
						{
							if ((inode->i_mode & 0xf000) == 0x4000)
							{
								memcpy(name, single_dir_del->name, single_dir_del->name_len);
								name[single_dir_del->name_len] = 0;
								printf("SUPER RICORSIONE NELLA DIR CANCELLATA %s INODE %d \n ",name, single_dir_del->inode);
								scan_dir(single_dir_del->inode, level + 1, in_n);
							}
							else
								printf("PULTROPPO NON E UNA DIR \n ");
						}
					}*/

				}

				i += 4;
			}	
		}	

		real_offset += single_dir->rec_len;
	
	//	printf("SHIFTO DI %d namelen %d reclen %d I %d\n  ",shift,single_dir->name_len, single_dir->rec_len, i);
	
	}

//	printf("situazione attuale %d \n",dir_blocks_n);

	free(dir);

	return 0;
}


int from_inode_number_to_path (int parent_in, int current_in, char *path)
{
	char path_temp[700];
	char path_temp2[700];
	struct ext2_inode *in;
	struct ext2_dir_entry_2 *dir;
	struct ext2_dir_entry_2 *dir_base;
	char *temp_p;
	
	int parent;

	memset(path_temp, 0, 700);
	memset(path_temp2, 0, 700);

	from_inode_number_to_inode(parent_in, &in);

	from_inode_to_dir(in, &dir_base);	
	
	dir = dir_base;

	while ((dir->name_len > 0) && (dir->rec_len > 0))
	{
		if (((dir->name_len & 0x00ff) == 2) && 
		    (dir->name[0] == '.'))
			parent = dir->inode;

		if (dir->inode == current_in)
		{
			int name_len = dir->name_len & 0x00ff;
			memcpy(path_temp, dir->name, name_len);
			path_temp[name_len] = 0;

			sprintf(path_temp2, "%s/%s", path_temp, path);
			memcpy(path, path_temp2, strlen(path_temp2) + 1);

			if ((parent == 2) && (parent_in == 2))
				return(0);
			else
				from_inode_number_to_path(parent, parent_in, path);
		}

		int shift = dir->rec_len;

		temp_p = dir;
		dir = &temp_p[shift];
	}

	free(dir_base);

	return 0;
}



int mkdir_birra(char *path)
{
	char temp_dir[256];

	int i; /* skippa il primo / */
	int z = 0;

	memset(temp_dir,0,256);

	temp_dir[0] = '/';

	for (i=1;i<strlen(path);i++)
	{
		if (path[i] == '/')
		{
			if (mkdir(temp_dir, 0755) == -1)
			{
				if (errno != EEXIST)
					return (1);
			}
		}
			
		temp_dir[i] = path[i];	
	}

	if (path[i - 1] != '/')
	{
		if (mkdir(temp_dir, 0755) == -1)
		{
			if (errno != EEXIST)
				return (1);
		}
	}

	return 0;
}



int check_dir(struct ext2_dir_entry_2 *dir)
{

	char name[300];
	int len;

//	dir_entry_dump(dir);

	if ((dir->rec_len > EXT2_BLOCK_SIZE(&super_block_s)) ||
	    (dir->rec_len % 4) ||
	    (dir->rec_len < 8) || 
	    (dir->inode > super_block_s.s_inodes_count) || 
	    (dir->inode < 0) || 
	    (dir->inode <= 0) || 
	    ((dir->name_len + 8) > dir->rec_len))
		return (1);

	return(0);
}



int set_iis(struct ext2_dir_entry_2 *dir, info_inodes *iis)
{
	int name_len;

	if (((dir->name_len & 0x00ff) == 1) && 
	    (dir->name[0] == '.'))
	{
		iis->current = dir->inode;
	}

	if (((dir->name_len & 0x00ff) == 2) && 
	    (dir->name[0] == '.'))
	{
		iis->parent = dir->inode;
	}

//	printf("dir inode %d iis inode %d \n ", dir->inode, iis->inode);

	if (dir->inode == iis->inode)
	{
		name_len = dir->name_len;
		memcpy(iis->name, dir->name, name_len);
		iis->name[name_len] = 0;

	}


	return (0);
}


int is_inode(int block)
{

	if ((block > group_desc_p[0].bg_inode_table) &&
	    (block < group_desc_p[groups_n -1].bg_inode_table))
		return (1);

	return(0);
}



int inode_is_allocated(int in_n)
{

	int i;

	int group = (in_n - 1) / EXT2_INODES_PER_GROUP(&super_block_s);

	unsigned int bit = (in_n - 1) - (group * EXT2_INODES_PER_GROUP(&super_block_s));

	if (group > groups_n)
		return(1);

	if (info_groups_p[group].inodes_p == 0)
	{
		if (allocated_groups > 30) 
		{
			for (i=0; (i<group && i<groups_n); i++)
			{
				if (info_groups_p[i].inodes_p != NULL)
				{
					free(info_groups_p[i].inodes_p);
					free(info_groups_p[i].inodes_bitmap_p);
					info_groups_p[i].inodes_p = 0;
					allocated_groups--;
				}
			}
		}

		if (load_group_inodes(group))
			return 1;
	}

	int member = bit / 32;
	int shift = bit % 32;

	bitmap_t temp_bitmap = info_groups_p[group].inodes_bitmap_p[member];

	return ((temp_bitmap >> shift) & 0x0001);
}	


int get_inode_from_descriptor_block(int *blocknr, journal_header_t *jh, char *journal, int blocksize, int search_blocknr, int inode_offset, struct ext2_inode **inode)
{
	journal_block_tag_t *jbt;
	int offset = sizeof(journal_header_t);
	char *temp = jh;
	int tag_block, tag_flags;
	int blocknr_temp;
	int temp_offset, i;
	int device_offset;
	int temp_inode;
	int name_len;
	char name[300];
	char *temp_p;
	int parent,current;
	char filename[700];
	char path[700];
	int shift;

	struct ext2_inode *in, *last_good_in = NULL;

	blocknr_temp = *blocknr;

	memset(name,0,300);
	
	++blocknr_temp;
	WRAP(journal_superblock, blocknr_temp);

	struct ext2_dir_entry_2 *dir = NULL;

	parent = 0;

	do  	
	{

		jbt = &temp[offset];

		offset += sizeof(journal_block_tag_t);

		if (offset > blocksize)
			break;
		
		tag_flags = be32_to_cpu(jbt->t_flags);		
		tag_block = be32_to_cpu(jbt->t_blocknr);		

		if (tag_block == search_blocknr)
		{
			temp_offset = blocknr_temp * blocksize;

			temp_offset += inode_offset;

			in = &journal[temp_offset];

			if ((in->i_dtime == 0) && (in->i_size > 0))
			{
				last_good_in = in;
			}
		}
			

		if (tag_flags & JFS_FLAG_DELETED)
			printf("il blocco %d e stato cancellato \n");

                if (!(tag_flags & JFS_FLAG_SAME_UUID))
                        offset += 16;

		++blocknr_temp;
		WRAP(journal_superblock, blocknr_temp);

	} while (!(tag_flags & JFS_FLAG_LAST_TAG));

	*blocknr = blocknr_temp;

	if (last_good_in)
	{
		*inode = last_good_in;
		return(0);
	}

	return(1);
}	



int get_inode_from_journal(int search_blocknr, int inode_offset, char *journal_p, struct ext2_inode **inode)
{
	journal_superblock_t *journal_superblock;
	journal_block_tag_t *journal_block_tag;
	journal_header_t *journal_header;

	journal_superblock = journal_p;

	int blocksize = be32_to_cpu(journal_superblock->s_blocksize);
	int blocknr = be32_to_cpu(journal_superblock->s_start);
	int transaction = be32_to_cpu(journal_superblock->s_sequence);

	while (1)
	{
		journal_header = &journal[blocknr * blocksize];

		if (journal_header == NULL)
		{
			printf("errore lettura journal \n");
			exit(1);
		}

		if ((be32_to_cpu(journal_header->h_sequence) != transaction) || (be32_to_cpu(journal_header->h_magic) != JFS_MAGIC_NUMBER))
			break;


	        switch (be32_to_cpu(journal_header->h_blocktype))
        	{
                	case JFS_DESCRIPTOR_BLOCK:
				if (get_inode_from_descriptor_block(&blocknr, journal_header, journal, be32_to_cpu(journal_superblock->s_blocksize), search_blocknr, inode_offset, inode) == 0)
					return(0);
                        	break;

                	case JFS_COMMIT_BLOCK:
				blocknr++;
				transaction++;
                        	break;

                	case JFS_REVOKE_BLOCK:
				blocknr++;
                        	break;

                	default:
                        	printf(" blocktype    : UNK \n");
        	}

	}	

	return 1;
}

